"""
.. module:: networks
   :synopsis: Flow neural networks for single speed and fast-slow hierarchies
.. moduleauthor:: Adam Moss <adam.moss@nottingham.ac.uk>
Coupling layer from https://github.com/ikostrikov/pytorch-flows modified 
to use variable number of layers
"""

import math

import torch
import torch.nn as nn


class BatchNormFlow(nn.Module):
    """ An implementation of a batch normalization layer from
    Density estimation using Real NVP
    (https://arxiv.org/abs/1605.08803).
    """

    def __init__(self, num_inputs, momentum=0.0, eps=1e-5):
        super(BatchNormFlow, self).__init__()

        self.log_gamma = nn.Parameter(torch.zeros(num_inputs))
        self.beta = nn.Parameter(torch.zeros(num_inputs))
        self.momentum = momentum
        self.eps = eps

        self.register_buffer('running_mean', torch.zeros(num_inputs))
        self.register_buffer('running_var', torch.ones(num_inputs))

    def forward(self, inputs, cond_inputs=None, mode='direct'):
        if mode == 'direct':
            if self.training:
                self.batch_mean = inputs.mean(0)
                self.batch_var = (
                    inputs - self.batch_mean).pow(2).mean(0) + self.eps

                self.running_mean.mul_(self.momentum)
                self.running_var.mul_(self.momentum)

                self.running_mean.add_(self.batch_mean.data *
                                       (1 - self.momentum))
                self.running_var.add_(self.batch_var.data *
                                      (1 - self.momentum))

                mean = self.batch_mean
                var = self.batch_var
            else:
                mean = self.running_mean
                var = self.running_var

            x_hat = (inputs - mean) / var.sqrt()
            y = torch.exp(self.log_gamma) * x_hat + self.beta
            return y, (self.log_gamma - 0.5 * torch.log(var)).sum(
                -1, keepdim=True)
        else:
            if self.training:
                mean = self.batch_mean
                var = self.batch_var
            else:
                mean = self.running_mean
                var = self.running_var

            x_hat = (inputs - self.beta) / torch.exp(self.log_gamma)

            y = x_hat * var.sqrt() + mean

            return y, (-self.log_gamma + 0.5 * torch.log(var)).sum(
                -1, keepdim=True)


class CouplingLayer(nn.Module):
    """ An implementation of a coupling layer
    from RealNVP (https://arxiv.org/abs/1605.08803).
    """

    def __init__(self,
                 num_inputs,
                 num_hidden,
                 mask,
                 num_cond_inputs=None,
                 s_act='tanh',
                 t_act='relu',
                 num_layers=2):
        super(CouplingLayer, self).__init__()

        self.num_inputs = num_inputs
        self.mask = mask

        activations = {'relu': nn.ReLU, 'sigmoid': nn.Sigmoid, 'tanh': nn.Tanh}
        s_act_func = activations[s_act]
        t_act_func = activations[t_act]

        if num_cond_inputs is not None:
            total_inputs = num_inputs + num_cond_inputs
        else:
            total_inputs = num_inputs
        
        scale_layers = [nn.Linear(total_inputs, num_hidden), s_act_func()]
        for i in range(0, num_layers):
            scale_layers += [nn.Linear(num_hidden, num_hidden), s_act_func()]
        scale_layers += [nn.Linear(num_hidden, num_inputs)]
        self.scale_net = nn.Sequential(*scale_layers)
        translate_layers = [nn.Linear(total_inputs, num_hidden), t_act_func()]
        for i in range(0, num_layers):
            translate_layers += [nn.Linear(num_hidden, num_hidden), t_act_func()]
        translate_layers += [nn.Linear(num_hidden, num_inputs)]
        self.translate_net = nn.Sequential(*translate_layers)
            
        def init(m):
            if isinstance(m, nn.Linear):
                m.bias.data.fill_(0)
                nn.init.orthogonal_(m.weight.data)

    def forward(self, inputs, cond_inputs=None, mode='direct'):
        mask = self.mask
        
        masked_inputs = inputs * mask
        if cond_inputs is not None:
            masked_inputs = torch.cat([masked_inputs, cond_inputs], -1)
        
        if mode == 'direct':
            log_s = self.scale_net(masked_inputs) * (1 - mask)
            t = self.translate_net(masked_inputs) * (1 - mask)
            s = torch.exp(log_s)
            return inputs * s + t, log_s.sum(-1, keepdim=True)
        else:
            log_s = self.scale_net(masked_inputs) * (1 - mask)
            t = self.translate_net(masked_inputs) * (1 - mask)
            s = torch.exp(-log_s)
            return (inputs - t) * s, -log_s.sum(-1, keepdim=True)


class FlowSequential(nn.Sequential):
    """ A sequential container for flows.
    In addition to a forward pass it implements a backward pass and
    computes log jacobians.
    """

    def forward(self, inputs, cond_inputs=None, mode='direct', logdets=None):
        """ Performs a forward or backward pass for flow modules.
        Args:
            inputs: a tuple of inputs and logdets
            mode: to run direct computation or inverse
        """
        self.num_inputs = inputs.size(-1)

        if logdets is None:
            logdets = torch.zeros(inputs.size(0), 1, device=inputs.device)

        assert mode in ['direct', 'inverse']
        if mode == 'direct':
            for module in self._modules.values():
                inputs, logdet = module(inputs, cond_inputs, mode)
                logdets += logdet
        else:
            for module in reversed(self._modules.values()):
                inputs, logdet = module(inputs, cond_inputs, mode)
                logdets += logdet

        return inputs, logdets

    def log_probs(self, inputs, cond_inputs=None):
        u, log_jacob = self(inputs, cond_inputs)
        log_probs = (-0.5 * u.pow(2) - 0.5 * math.log(2 * math.pi)).sum(
            -1, keepdim=True)
        return (log_probs + log_jacob).sum(-1, keepdim=True)

    def sample(self, num_samples=None, noise=None, cond_inputs=None):
        if noise is None:
            noise = torch.Tensor(num_samples, self.num_inputs).normal_()
        device = next(self.parameters()).device
        noise = noise.to(device)
        if cond_inputs is not None:
            cond_inputs = cond_inputs.to(device)
        samples = self.forward(noise, cond_inputs, mode='inverse')[0]
        return samples


class SingleSpeed(nn.Module):

    def __init__(self, num_inputs, num_hidden, num_blocks, num_layers):
        super(SingleSpeed, self).__init__()

        mask = torch.arange(0, num_inputs) % 2
        mask = mask.float()
        modules = []
        for _ in range(num_blocks):
            modules += [
                CouplingLayer(
                    num_inputs, num_hidden, mask, None,
                    s_act='tanh', t_act='relu', num_layers=num_layers)
            ]
            mask = 1 - mask
        self.net = FlowSequential(*modules)

    def forward(self, inputs, cond_inputs=None, mode='direct', logdets=None):
        return self.net(inputs, cond_inputs=cond_inputs, mode=mode, logdets=logdets)

    def log_probs(self, inputs, cond_inputs=None):
        return self.net.log_probs(inputs, cond_inputs=None)

    def sample(self, num_samples=None, noise=None, cond_inputs=None):
        return self.net.sample(num_samples=num_samples, noise=noise, cond_inputs=cond_inputs)


class FastSlow(nn.Module):

    def __init__(self, num_fast, num_slow, num_hidden, num_blocks, num_layers):
        super(FastSlow, self).__init__()

        self.num_fast = num_fast
        self.num_slow = num_slow
        self.num_inputs = num_fast + num_slow

        # Fast block
        mask_fast = torch.arange(0, num_fast) % 2
        mask_fast = mask_fast.float()
        modules_fast = []
        for _ in range(num_blocks):
            modules_fast += [
                CouplingLayer(
                    num_fast, num_hidden, mask_fast, None,
                    s_act='tanh', t_act='relu', num_layers=num_layers)
            ]
            mask_fast = 1 - mask_fast
        self.net_fast = FlowSequential(*modules_fast)

        # Slow block
        mask_slow = torch.arange(0, num_slow) % 2
        mask_slow = mask_slow.float()
        modules_slow = []
        for _ in range(num_blocks):
            modules_slow += [
                CouplingLayer(
                    num_slow, num_hidden, mask_slow, None,
                    s_act='tanh', t_act='relu', num_layers=num_layers)
            ]
            mask_slow = 1 - mask_slow
        self.net_slow = FlowSequential(*modules_slow)

        # Combine fast and slow such that slow is unnchanged just by updating fast block
        modules = []
        mask = torch.cat((torch.ones(num_slow), torch.zeros(num_fast)))
        modules = [
            CouplingLayer(
                num_slow + num_fast, num_hidden, mask, None,
                s_act='tanh', t_act='relu', num_layers=num_layers)
        ]
        self.net = FlowSequential(*modules)

    def forward(self, inputs, cond_inputs=None, mode='direct', logdets=None):
        assert mode in ['direct', 'inverse']
        if mode == 'direct':
            slow, logdets_slow = self.net_slow(inputs[:, 0:self.num_slow], mode=mode)
            fast, logdets_fast = self.net_fast(inputs[:, self.num_slow:self.num_slow+self.num_fast], mode=mode)
            inputs = torch.cat((slow, fast), dim=1)
            inputs, logdets = self.net(inputs, mode=mode)
        else:
            inputs, logdets = self.net(inputs, mode=mode)
            slow, logdets_slow = self.net_slow(inputs[:, 0:self.num_slow], mode=mode)
            fast, logdets_fast = self.net_fast(inputs[:, self.num_slow:self.num_slow+self.num_fast], mode=mode)
            inputs = torch.cat((slow, fast), dim=1)
        return inputs, logdets_slow + logdets_fast + logdets

    def log_probs(self, inputs, cond_inputs=None):
        slow, logdets_slow = self.net_slow(inputs[:, 0:self.num_slow])
        fast, logdets_fast = self.net_fast(inputs[:, self.num_slow:self.num_slow+self.num_fast])
        inputs = torch.cat((slow, fast), dim=1)
        u, log_jacob = self.net(inputs)
        log_probs = (-0.5 * u.pow(2) - 0.5 * math.log(2 * math.pi)).sum(
            -1, keepdim=True)
        return (log_probs + log_jacob + logdets_slow + logdets_fast).sum(-1, keepdim=True)

    def sample(self, num_samples=None, noise=None, cond_inputs=None):
        if noise is None:
            noise = torch.Tensor(num_samples, self.num_inputs).normal_()
        device = next(self.parameters()).device
        noise = noise.to(device)
        if cond_inputs is not None:
            cond_inputs = cond_inputs.to(device)
        samples = self.forward(noise, cond_inputs, mode='inverse')[0]
        return samples



import logging
import sys
import os
import numpy as np
import pandas as pd
import errno


def create_logger(module_name, level=logging.INFO):
    logger = logging.getLogger(module_name)
    logger.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    formatter = logging.Formatter('[{}] [%(levelname)s] %(message)s'.format(module_name))
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def save_ess(ess, path):
    df = pd.DataFrame(np.reshape(ess, [1, -1]))
    df.to_csv('{}/ess.csv'.format(path), mode='a', header=False)


def ensure_directory(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def make_run_dir(log_dir, run_num=None, append_run_num=True):
    """Generates a new numbered directory for this run to store output"""
    try:
        os.makedirs(log_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    if run_num is None or run_num == '':
        run_num = (sum(os.path.isdir(os.path.join(log_dir,i))
                      for i in os.listdir(log_dir)) + 1)
    if append_run_num:
        run_dir = os.path.join(log_dir, 'run%s' % run_num)
    else:
        run_dir = log_dir
    if not os.path.isdir(run_dir):
        print('Creating directory for new run %s' % run_dir)
        os.makedirs(run_dir)
    if not os.path.isdir(os.path.join(run_dir, 'info')):
        os.makedirs(os.path.join(run_dir, 'info'))
        os.makedirs(os.path.join(run_dir, 'results'))
        os.makedirs(os.path.join(run_dir, 'chains'))
        os.makedirs(os.path.join(run_dir, 'extra'))
        os.makedirs(os.path.join(run_dir, 'plots'))

    return {'run_dir': run_dir,
            'info': os.path.join(run_dir, 'info'),
            'results': os.path.join(run_dir, 'results'),
            'chains': os.path.join(run_dir, 'chains'),
            'extra': os.path.join(run_dir, 'extra'),
            'plots': os.path.join(run_dir, 'plots')
            }



"""
.. module:: trainer
   :synopsis: Train the flow and perform sampling in latent space
.. moduleauthor:: Adam Moss <adam.moss@nottingham.ac.uk>
"""

from __future__ import print_function
from __future__ import division

import os
import time
import copy

import torch
from torch.utils.data import DataLoader
import numpy as np
from tensorboardX import SummaryWriter
from sklearn.model_selection import train_test_split
import scipy.spatial
from tqdm import tqdm

import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt

from nnest.networks import SingleSpeed, FastSlow, BatchNormFlow
from nnest.utils.logger import create_logger


class Trainer(object):

    def __init__(self,
                 xdim,
                 ndim,
                 nslow=0,
                 batch_size=100,
                 flow='nvp',
                 num_blocks=5,
                 num_layers=2,
                 oversample_rate=-1,
                 train=True,
                 load_model='',
                 log_dir='logs',
                 use_gpu=False,
                 log=True
                 ):

        self.device = torch.device(
            'cuda' if use_gpu and torch.cuda.is_available() else 'cpu')

        self.x_dim = xdim
        self.z_dim = xdim

        self.batch_size = batch_size
        self.total_iters = 0

        assert xdim > nslow
        nfast = xdim - nslow
        self.nslow = nslow

        if oversample_rate > 0:
            self.oversample_rate = oversample_rate
        else:
            self.oversample_rate = nfast / xdim

        if flow.lower() == 'nvp':
            if nslow > 0:
                self.netG = FastSlow(nfast, nslow, ndim, num_blocks, num_layers)
            else:
                self.netG = SingleSpeed(xdim, ndim, num_blocks, num_layers)
        else:
            raise NotImplementedError

        if train and not load_model:
            if log_dir is not None:
                self.path = log_dir
                if not os.path.exists(os.path.join(self.path, 'models')):
                    os.makedirs(os.path.join(self.path, 'models'))
                if not os.path.exists(os.path.join(self.path, 'data')):
                    os.makedirs(os.path.join(self.path, 'data'))
                if not os.path.exists(os.path.join(self.path, 'chains')):
                    os.makedirs(os.path.join(self.path, 'chains'))
                if not os.path.exists(os.path.join(self.path, 'plots')):
                    os.makedirs(os.path.join(self.path, 'plots'))
            else:
                self.path = None
        else:
            self.path = os.path.join(log_dir, load_model)
            self.netG.load_state_dict(torch.load(
                os.path.join(self.path, 'models', 'netG.pt')
            ))

        self.optimizer = torch.optim.Adam(
            self.netG.parameters(), lr=0.0001, weight_decay=1e-6)

        self.logger = create_logger(__name__)
        self.log = log

        if self.path is not None:
            self.logger.info(self.netG)
            self.writer = SummaryWriter(self.path)

    def train(
            self,
            samples,
            max_iters=5000,
            log_interval=50,
            save_interval=50,
            noise=0.0,
            validation_fraction=0.1):

        start_time = time.time()

        if self.path:
            fig, ax = plt.subplots()
            ax.scatter(samples[:, 0], samples[:, 1])
            self.writer.add_figure('originals', fig, self.total_iters)
            np.save(
                os.path.join(self.path, 'data', 'originals.npy'),
                samples)

        if noise < 0:
            kdt = scipy.spatial.cKDTree(samples)
            dists, neighs = kdt.query(samples, 2)
            training_noise = .2 * np.mean(dists)
        else:
            training_noise = noise

        if self.log:
            self.logger.info('Number of training samples [%d]' % samples.shape[0])
            self.logger.info('Training noise [%5.4f]' % training_noise)

        X_train, X_valid = train_test_split(
            samples, test_size=validation_fraction)

        train_tensor = torch.from_numpy(X_train.astype(np.float32))
        train_dataset = torch.utils.data.TensorDataset(train_tensor)
        train_loader = torch.utils.data.DataLoader(
            train_dataset, batch_size=self.batch_size, shuffle=True)

        valid_tensor = torch.from_numpy(X_valid.astype(np.float32))
        valid_dataset = torch.utils.data.TensorDataset(valid_tensor)
        valid_loader = torch.utils.data.DataLoader(
            valid_dataset, batch_size=X_valid.shape[0], shuffle=False, drop_last=True)

        best_validation_loss = float('inf')
        best_validation_epoch = 0
        best_model = copy.deepcopy(self.netG)

        for epoch in range(1, max_iters + 1):

            self.total_iters += 1

            train_loss = self._train(
                epoch, self.netG, train_loader, noise=training_noise)
            validation_loss = self._validate(epoch, self.netG, valid_loader)

            if validation_loss < best_validation_loss:
                best_validation_epoch = epoch
                best_validation_loss = validation_loss
                best_model = copy.deepcopy(self.netG)

            if epoch == 1 or epoch % log_interval == 0:
                self.logger.info('Epoch [%i] train loss [%5.4f] validation loss [%5.4f]' % (
                    epoch, train_loss, validation_loss))

            if self.path:
                self.writer.add_scalar('loss', validation_loss, self.total_iters)
                if epoch % save_interval == 0:
                    torch.save(
                        self.netG.state_dict(),
                        os.path.join(self.path, 'models', 'netG.pt')
                    )
                    self._train_plot(self.netG, samples)

        self.logger.info('Best epoch [%i] validation loss [%5.4f]' % (best_validation_epoch, best_validation_loss))

        self.netG.load_state_dict(best_model.state_dict())

    def sample(
            self,
            mcmc_steps=20,
            alpha=1.0,
            dynamic=True,
            batch_size=1,
            loglike=None,
            init_x=None,
            logl=None,
            loglstar=None,
            transform=None,
            show_progress=False,
            plot=False,
            out_chain=None,
            max_prior=None,
            max_start_tries=100):

        self.netG.eval()

        samples = []
        latent = []
        likes = []

        if transform is None:
            def transform(x): return x

        if init_x is not None:
            batch_size = init_x.shape[0]
            z, _ = self.netG(torch.from_numpy(init_x).float().to(self.device))
            z = z.detach()
            # Add the backward version of x rather than init_x due to numerical precision
            x, _ = self.netG(z, mode='inverse')
            x = x.detach().cpu().numpy()
            if logl is None:
                logl = loglike(transform(x))
        else:
            if logl is None:
                for i in range(max_start_tries):
                    z = torch.randn(batch_size, self.z_dim, device=self.device)
                    x, _ = self.netG(z, mode='inverse')
                    x = x.detach().cpu().numpy()
                    logl = loglike(transform(x))
                    if np.all(logl > -1e30):
                        break
                    if i == max_start_tries - 1:
                        raise Exception('Could not find starting value')
            else:
                z = torch.randn(batch_size, self.z_dim, device=self.device)
                x, _ = self.netG(z, mode='inverse')
                x = x.detach().cpu().numpy()
                logl = loglike(transform(x))                   

        samples.append(x)
        likes.append(logl)

        iters = range(mcmc_steps)
        if show_progress:
            iters = tqdm(iters)

        scale = alpha
        accept = 0
        reject = 0
        ncall = 0

        if out_chain is not None:
            if batch_size == 1:
                files = [open(out_chain + '.txt', 'w')]
            else:
                files = [open(out_chain + '_%s.txt' % (ib + 1), 'w') for ib in range(batch_size)]

        for i in iters:

            dz = torch.randn_like(z) * scale
            if self.nslow > 0 and np.random.uniform() < self.oversample_rate:
                fast = True
                dz[:, 0:self.nslow] = 0.0
            else:
                fast = False
            z_prime = z + dz

            # Jacobian is det d f^{-1} (z)/dz
            x, log_det_J = self.netG(z, mode='inverse')
            x_prime, log_det_J_prime = self.netG(z_prime, mode='inverse')
            x = x.detach().cpu().numpy()
            x_prime = x_prime.detach().cpu().numpy()
            delta_log_det_J = (log_det_J_prime - log_det_J).detach()
            log_ratio_1 = delta_log_det_J.squeeze(dim=1)

            # Check not out of prior range
            if max_prior is not None:
                prior = np.logical_or(
                    np.abs(x) > max_prior,
                    np.abs(x_prime) > max_prior)
                idx = np.where([np.any(p) for p in prior])
                log_ratio_1[idx] = -np.inf

            rnd_u = torch.rand(log_ratio_1.shape, device=self.device)
            ratio = log_ratio_1.exp().clamp(max=1)
            mask = (rnd_u < ratio).int()
            logl_prime = np.full(batch_size, logl)

            # Only evaluate likelihood if prior and volume is accepted
            if loglike is not None and transform is not None:
                for idx, im in enumerate(mask):
                    if im:
                        if not fast:
                            ncall += 1
                        lp = loglike(transform(x_prime[idx]))
                        if loglstar is not None:
                            if np.isfinite(lp) and lp >= loglstar:
                                logl_prime[idx] = lp
                            else:
                                mask[idx] = 0
                        else:
                            if lp >= logl[idx]:
                                logl_prime[idx] = lp
                            elif rnd_u[idx].cpu().numpy() < np.clip(np.exp(lp - logl[idx]), 0, 1):
                                logl_prime[idx] = lp
                            else:
                                mask[idx] = 0

            accept += torch.sum(mask).cpu().numpy()
            reject += batch_size - torch.sum(mask).cpu().numpy()

            if dynamic:
                if accept > reject:
                    scale *= np.exp(1. / accept)
                if accept < reject:
                    scale /= np.exp(1. / reject)

            m = mask[:, None].float()
            z = (z_prime * m + z * (1 - m)).detach()

            m = mask
            logl = logl_prime * m.cpu().numpy() + logl * (1 - m.cpu().numpy())

            x, _ = self.netG(z, mode='inverse')
            x = x.detach().cpu().numpy()
            samples.append(x)
            likes.append(logl)
            latent.append(z.cpu().numpy())

            if out_chain is not None:
                v = transform(x)
                for ib in range(batch_size):
                    files[ib].write("%.5E " % 1)
                    files[ib].write("%.5E " % -logl[ib])
                    files[ib].write(" ".join(["%.5E" % vi for vi in v[ib]]))
                    files[ib].write("\n")
                    files[ib].flush()

        # Transpose so shape is (chain_num, iteration, dim)
        samples = np.transpose(np.array(samples), axes=[1, 0, 2])
        likes = np.transpose(np.array(likes), axes=[1, 0])
        latent = np.transpose(np.array(latent), axes=[1, 0, 2])

        if self.path and plot:
            cmap = plt.cm.jet
            cmap.set_under('w', 1)
            fig, ax = plt.subplots()
            ax.hist2d(samples[:, -1, 0], samples[:, -1, 1],
                      bins=200, cmap=cmap, vmin=1, alpha=0.2)
            if self.writer is not None:
                self.writer.add_figure('chain', fig, self.total_iters)

        if out_chain is not None:
            for ib in range(batch_size):
                files[ib].close()

        return samples, likes, latent, scale, ncall

    def _jacobian(self, z):
        """ Calculate det d f^{-1} (z)/dz
        """
        z.requires_grad_(True)
        x, _ = self.netG(z, mode='inverse')
        J = torch.stack([torch.autograd.grad(outputs=x[:, i], inputs=z, retain_graph=True,
                                             grad_outputs=torch.ones(z.shape[0]))[0] for i in
                         range(x.shape[1])]).permute(1, 0, 2)
        return torch.stack([torch.log(torch.abs(torch.det(J[i, :, :])))
                            for i in range(x.shape[0])])

    def _train(self, epoch, model, loader, noise=0.0):

        model.train()
        train_loss = 0

        for batch_idx, data in enumerate(loader):
            if isinstance(data, list):
                if len(data) > 1:
                    cond_data = data[1].float()
                    cond_data = cond_data.to(self.device)
                else:
                    cond_data = None
                data = data[0]
            data = (data + noise * torch.randn_like(data)).to(self.device)
            self.optimizer.zero_grad()
            loss = -model.log_probs(data, cond_data).mean()
            train_loss += loss.item()
            loss.backward()
            self.optimizer.step()

        for module in model.modules():
            if isinstance(module, BatchNormFlow):
                module.momentum = 0

        with torch.no_grad():
            model(loader.dataset.tensors[0].to(data.device))

        for module in model.modules():
            if isinstance(module, BatchNormFlow):
                module.momentum = 1

        return train_loss / len(loader.dataset)

    def _validate(self, epoch, model, loader):

        model.eval()
        val_loss = 0
        cond_data = None

        for batch_idx, data in enumerate(loader):
            if isinstance(data, list):
                if len(data) > 1:
                    cond_data = data[1].float()
                    cond_data = cond_data.to(self.device)
                data = data[0]
            data = data.to(self.device)
            with torch.no_grad():
                # sum up batch loss
                val_loss += -model.log_probs(data, cond_data).mean().item()

        return val_loss / len(loader.dataset)

    def _train_plot(self, model, dataset, plot_grid=True):

        model.eval()
        with torch.no_grad():
            x_synth = model.sample(dataset.size).detach().cpu().numpy()
            x_data = torch.from_numpy(dataset).float()
            z, _ = model(x_data)
            z = z.detach().cpu().numpy()
            if plot_grid and self.x_dim == 2:
                grid = []
                for x in np.linspace(np.min(dataset[:, 0]), np.max(dataset[:, 0]), 10):
                    for y in np.linspace(np.min(dataset[:, 1]), np.max(dataset[:, 1]), 5000):
                        grid.append([x, y])
                for y in np.linspace(np.min(dataset[:, 1]), np.max(dataset[:, 1]), 10):
                    for x in np.linspace(np.min(dataset[:, 0]), np.max(dataset[:, 0]), 5000):
                        grid.append([x, y])
                grid = np.array(grid)
                x_grid = torch.from_numpy(grid).float()
                z_grid, _ = model(x_grid)
                z_grid = z_grid.detach().cpu().numpy()

        if self.writer is not None:
            fig, ax = plt.subplots(2, figsize=(5, 10))
            ax[0].scatter(dataset[:, 0], dataset[:, 1], c=dataset[:, 0], s=4)
            ax[1].scatter(z[:, 0], z[:, 1], c=dataset[:, 0], s=4)
            self.writer.add_figure('latent', fig, self.total_iters)

        if self.path:
            fig, ax = plt.subplots(1, 3, figsize=(10, 4))
            if plot_grid and self.x_dim == 2:
                ax[0].scatter(grid[:, 0], grid[:, 1], c=grid[:, 0], marker='.', s=1, linewidths=0)
            ax[0].scatter(dataset[:, 0], dataset[:, 1], s=4)
            ax[0].set_title('Real data')
            if plot_grid and self.x_dim == 2:
                ax[1].scatter(z_grid[:, 0], z_grid[:, 1], c=grid[:, 0], marker='.', s=1, linewidths=0)
            ax[1].scatter(z[:, 0], z[:, 1], s=4)
            ax[1].set_title('Latent data')
            ax[1].set_xlim([-3, 3])
            ax[1].set_ylim([-3, 3])
            ax[2].scatter(x_synth[:, 0], x_synth[:, 1], s=2)
            ax[2].set_title('Synthetic data')
            plt.tight_layout()
            plt.savefig(os.path.join(self.path, 'plots', 'plot_%s.png' % self.total_iters))
            plt.close()



"""
.. module:: nested
   :synopsis: Nested sampler
.. moduleauthor:: Adam Moss <adam.moss@nottingham.ac.uk>
Performs nested sampling to calculate the Bayesian evidence and posterior samples
Some parts are from the Nestle library by Kyle Barbary (https://github.com/kbarbary/nestle)
"""

from __future__ import print_function
from __future__ import division

import os
import csv

import numpy as np

from nnest.sampler import Sampler


class NestedSampler(Sampler):

    def __init__(self,
                 x_dim,
                 loglike,
                 transform=None,
                 append_run_num=True,
                 run_num=None,
                 hidden_dim=128,
                 num_slow=0,
                 num_derived=0,
                 batch_size=100,
                 flow='nvp',
                 num_blocks=5,
                 num_layers=2,
                 log_dir='logs/test',
                 num_live_points=1000
                 ):

        self.num_live_points = num_live_points
        self.sampler = 'nested'

        super(NestedSampler, self).__init__(x_dim, loglike, transform=transform, append_run_num=append_run_num,
                                            run_num=run_num, hidden_dim=hidden_dim, num_slow=num_slow, 
                                            num_derived=num_derived, batch_size=batch_size, flow=flow,
                                            num_blocks=num_blocks, num_layers=num_layers, log_dir=log_dir)

        if self.log:
            self.logger.info('Num live points [%d]' % (self.num_live_points))

        if self.log:
            with open(os.path.join(self.logs['results'], 'results.csv'), 'w') as f:
                writer = csv.writer(f)
                writer.writerow(['step', 'acceptance', 'min_ess',
                                 'max_ess', 'jump_distance', 'scale', 'loglstar', 'logz', 'fraction_remain', 'ncall'])

    def run(
            self,
            mcmc_steps=0,
            mcmc_burn_in=0,
            mcmc_batch_size=10,
            max_iters=1000000,
            update_interval=None,
            log_interval=None,
            dlogz=0.5,
            train_iters=50,
            volume_switch=0,
            alpha=0.0,
            noise=-1.0,
            num_test_samples=0,
            test_mcmc_steps=1000,
            test_mcmc_burn_in=0):

        if update_interval is None:
            update_interval = max(1, round(self.num_live_points))
        else:
            update_interval = round(update_interval)
            if update_interval < 1:
                raise ValueError("update_interval must be >= 1")

        if log_interval is None:
            log_interval = max(1, round(0.2 * self.num_live_points))
        else:
            log_interval = round(log_interval)
            if log_interval < 1:
                raise ValueError("log_interval must be >= 1")

        if mcmc_steps <= 0:
            mcmc_steps = 5 * self.x_dim

        if volume_switch <= 0:
            volume_switch = 1 / mcmc_steps

        if alpha == 0.0:
            alpha = 2 / self.x_dim ** 0.5

        if self.log:
            self.logger.info('MCMC steps [%d] alpha [%5.4f] volume switch [%5.4f]' % (mcmc_steps, alpha, volume_switch))

        if self.use_mpi:
            self.logger.info('Using MPI with rank [%d]' % (self.mpi_rank))
            if self.mpi_rank == 0:
                active_u = 2 * (np.random.uniform(size=(self.num_live_points, self.x_dim)) - 0.5)
            else:
                active_u = np.empty((self.num_live_points, self.x_dim), dtype=np.float64)
            self.comm.Bcast(active_u, root=0)
        else:
            active_u = 2 * (np.random.uniform(size=(self.num_live_points, self.x_dim)) - 0.5)
        active_v = self.transform(active_u)

        if self.use_mpi:
            if self.mpi_rank == 0:
                chunks = [[] for _ in range(self.mpi_size)]
                for i, chunk in enumerate(active_v):
                    chunks[i % self.mpi_size].append(chunk)
            else:
                chunks = None
            data = self.comm.scatter(chunks, root=0)
            active_logl = self.loglike(data)
            recv_active_logl = self.comm.gather(active_logl, root=0)
            recv_active_logl = self.comm.bcast(recv_active_logl, root=0)
            active_logl = np.concatenate(recv_active_logl, axis=0)
        else:
            active_logl = self.loglike(active_v)

        saved_v = []  # Stored points for posterior results
        saved_logl = []
        saved_logwt = []
        h = 0.0  # Information, initially 0.
        logz = -1e300  # ln(Evidence Z), initially Z=0
        logvol = np.log(1.0 - np.exp(-1.0 / self.num_live_points))
        fraction_remain = 1.0
        ncall = self.num_live_points  # number of calls we already made
        first_time = True
        nb = self.mpi_size * mcmc_batch_size

        for it in range(0, max_iters):

            # Worst object in collection and its weight (= volume * likelihood)
            worst = np.argmin(active_logl)
            logwt = logvol + active_logl[worst]

            # Update evidence Z and information h.
            logz_new = np.logaddexp(logz, logwt)
            h = (np.exp(logwt - logz_new) * active_logl[worst] + np.exp(logz - logz_new) * (h + logz) - logz_new)
            logz = logz_new

            # Add worst object to samples.
            saved_v.append(np.array(active_v[worst]))
            saved_logwt.append(logwt)
            saved_logl.append(active_logl[worst])

            expected_vol = np.exp(-it / self.num_live_points)

            # The new likelihood constraint is that of the worst object.
            loglstar = active_logl[worst]

            if expected_vol > volume_switch:

                nc = 0
                # Simple rejection sampling over prior
                while True:
                    u = 2 * (np.random.uniform(size=(1, self.x_dim)) - 0.5)
                    v = self.transform(u)
                    logl = self.loglike(v)
                    nc += 1
                    if logl > loglstar:
                        break
                if self.use_mpi:
                    recv_samples = self.comm.gather(u, root=0)
                    recv_likes = self.comm.gather(logl, root=0)
                    recv_nc = self.comm.gather(nc, root=0)
                    recv_samples = self.comm.bcast(recv_samples, root=0)
                    recv_likes = self.comm.bcast(recv_likes, root=0)
                    recv_nc = self.comm.bcast(recv_nc, root=0)
                    samples = np.concatenate(recv_samples, axis=0)
                    likes = np.concatenate(recv_likes, axis=0)
                    ncall += sum(recv_nc)
                else:
                    samples = np.array(u)
                    likes = np.array(logl)
                    ncall += nc
                for ib in range(0, self.mpi_size):
                    active_u[worst] = samples[ib, :]
                    active_v[worst] = self.transform(active_u[worst])
                    active_logl[worst] = likes[ib]

                if it % log_interval == 0 and self.log:
                    self.logger.info(
                        'Step [%d] loglstar [%5.4f] max logl [%5.4f] logz [%5.4f] vol [%6.5f] ncalls [%d]]' %
                        (it, loglstar, np.max(active_logl), logz, expected_vol, ncall))

            else:

                # MCMC
                if first_time or it % update_interval == 0:
                    self.trainer.train(active_u, max_iters=train_iters, noise=noise)
                    if num_test_samples > 0:
                        # Test multiple chains from worst point to check mixing
                        init_x = np.concatenate(
                            [active_u[worst:worst + 1, :] for i in range(num_test_samples)])
                        logl = np.concatenate(
                            [active_logl[worst:worst + 1] for i in range(num_test_samples)])
                        test_samples, _, _, scale, _ = self.trainer.sample(
                            loglike=self.loglike, init_x=init_x, logl=logl, loglstar=loglstar,
                            transform=self.transform, mcmc_steps=test_mcmc_steps + test_mcmc_burn_in,
                            max_prior=1, alpha=alpha)
                        np.save(os.path.join(self.logs['chains'], 'test_samples.npy'), test_samples)
                        self._chain_stats(test_samples, mean=np.mean(active_u, axis=0), std=np.std(active_u, axis=0))
                    first_time = False

                accept = False
                while not accept:
                    if nb == self.mpi_size * mcmc_batch_size:
                        # Get a new batch of trial points
                        nb = 0
                        idx = np.random.randint(
                            low=0, high=self.num_live_points, size=mcmc_batch_size)
                        init_x = active_u[idx, :]
                        logl = active_logl[idx]
                        samples, likes, latent, scale, nc = self.trainer.sample(
                            loglike=self.loglike, init_x=init_x, logl=logl, loglstar=loglstar,
                            transform=self.transform, mcmc_steps=mcmc_steps + mcmc_burn_in,
                            max_prior=1, alpha=alpha)
                        if self.use_mpi:
                            recv_samples = self.comm.gather(samples, root=0)
                            recv_likes = self.comm.gather(likes, root=0)
                            recv_nc = self.comm.gather(nc, root=0)
                            recv_samples = self.comm.bcast(recv_samples, root=0)
                            recv_likes = self.comm.bcast(recv_likes, root=0)
                            recv_nc = self.comm.bcast(recv_nc, root=0)
                            samples = np.concatenate(recv_samples, axis=0)
                            likes = np.concatenate(recv_likes, axis=0)
                            ncall += sum(recv_nc)
                        else:
                            ncall += nc
                    for ib in range(nb, self.mpi_size * mcmc_batch_size):
                        nb += 1
                        if np.all(samples[ib, 0, :] != samples[ib, -1, :]) and likes[ib, -1] > loglstar:
                            active_u[worst] = samples[ib, -1, :]
                            active_v[worst] = self.transform(active_u[worst])
                            active_logl[worst] = likes[ib, -1]
                            accept = True
                            break

                if it % log_interval == 0 and self.log:
                    acceptance, ess, jump_distance = self._chain_stats(samples, mean=np.mean(active_u, axis=0),
                                                                       std=np.std(active_u, axis=0))
                    np.save(
                        os.path.join(
                            self.logs['extra'],
                            'active_%s.npy' %
                            it),
                        active_v)
                    self.logger.info(
                        'Step [%d] loglstar [%5.4f] maxlogl [%5.4f] logz [%5.4f] vol [%6.5f] ncalls [%d] scale [%5.4f]' %
                        (it, loglstar, np.max(active_logl), logz, expected_vol, ncall, scale))
                    with open(os.path.join(self.logs['results'], 'results.csv'), 'a') as f:
                        writer = csv.writer(f)
                        writer.writerow([it, acceptance, np.min(ess), np.max(
                            ess), jump_distance, scale, loglstar, logz, fraction_remain, ncall])
                    self._save_samples(np.array(saved_v), np.exp(np.array(saved_logwt) - logz), np.array(saved_logl))

            # Shrink interval
            logvol -= 1.0 / self.num_live_points
            logz_remain = np.max(active_logl) - it / self.num_live_points
            fraction_remain = np.logaddexp(logz, logz_remain) - logz

            if self.log:
                self.trainer.writer.add_scalar('logz', logz, it)

            # Stopping criterion
            if fraction_remain < dlogz:
                break

        logvol = -len(saved_v) / self.num_live_points - np.log(self.num_live_points)
        for i in range(self.num_live_points):
            logwt = logvol + active_logl[i]
            logz_new = np.logaddexp(logz, logwt)
            h = (np.exp(logwt - logz_new) * active_logl[i] + np.exp(logz - logz_new) * (h + logz) - logz_new)
            logz = logz_new
            saved_v.append(np.array(active_v[i]))
            saved_logwt.append(logwt)
            saved_logl.append(active_logl[i])

        if self.log:
            with open(os.path.join(self.logs['results'], 'final.csv'), 'w') as f:
                writer = csv.writer(f)
                writer.writerow(['niter', 'ncall', 'logz', 'logzerr', 'h'])
                writer.writerow([it + 1, ncall, logz, np.sqrt(h / self.num_live_points), h])
            self._save_samples(np.array(saved_v), np.exp(np.array(saved_logwt) - logz), np.array(saved_logl))

        print("niter: {:d}\n ncall: {:d}\n nsamples: {:d}\n logz: {:6.3f} +/- {:6.3f}\n h: {:6.3f}"
              .format(it + 1, ncall, len(np.array(saved_v)), logz, np.sqrt(h / self.num_live_points), h))



from __future__ import division

import numpy as np


def auto_correlation_time(x, s, mu, var):
    b, t, d = x.shape
    act_ = np.zeros([d])
    for i in range(0, b):
        y = x[i] - mu
        p, n = y[:-s], y[s:]
        act_ += np.mean(p * n, axis=0) / var
    act_ = act_ / b
    return act_


def effective_sample_size(x, mu, var):
    """
    Calculate the effective sample size of sequence generated by MCMC.
    :param x:
    :param mu: mean of the variable
    :param var: variance of the variable
    :param logger: logg
    :return: effective sample size of the sequence
    Make sure that `mu` and `var` are correct!
    """
    # batch size, step, dimension
    b, t, d = x.shape
    ess_ = np.ones([d])
    for s in range(1, t):
        p = auto_correlation_time(x, s, mu, var)
        if np.sum(p > 0.05) == 0:
            break
        else:
            for j in range(0, d):
                if p[j] > 0.05:
                    ess_[j] += 2.0 * p[j] * (1.0 - float(s) / t)

    return t / ess_


def acceptance_rate(z):
    cnt = z.shape[0] * (z.shape[1] - 1)
    for i in range(0, z.shape[0]):
        for j in range(1, z.shape[1]):
            if np.min(np.equal(z[i, j - 1], z[i, j])):
                cnt -= 1
    return cnt / float(z.shape[0] * (z.shape[1] - 1))


def mean_jump_distance(z):
    d = 0
    cnt = z.shape[0] * (z.shape[1] - 1)
    for i in range(0, z.shape[0]):
        for j in range(1, z.shape[1]):
            d += np.linalg.norm(z[i, j - 1] - z[i, j])
    return d / cnt


def gelman_rubin_diagnostic(x, mu=None):
    m, n = x.shape[0], x.shape[1]
    theta = np.mean(x, axis=1)
    sigma = np.var(x, axis=1)
    # theta_m = np.mean(theta, axis=0)
    theta_m = mu if mu is not None else np.mean(theta, axis=0)
    b = float(n) / float(m-1) * np.sum((theta - theta_m) ** 2)
    w = 1. / (float(m) * np.sum(sigma, axis=0) + 1e-5)
    v = float(n-1) / float(n) * w + float(m+1) / float(m * n) * b
    r_hat = np.sqrt(v / w)
    return r_hat


def autocor_ESS(A):
    A = A * (A > 0.05)
    return 1. / (1. + 2 * np.sum(A[1:]))


def autocovariance(X, tau=0):
  dT, dN, dX = np.shape(X)
  s = 0.
  for t in range(dT - tau):
    x1 = X[t, :, :]
    x2 = X[t+tau, :, :]

    s += np.sum(x1 * x2) / dN

  return s / (dT - tau)


def acl_spectrum(X, scale):
    n = X.shape[0]
    return np.array([autocovariance(X / scale, tau=t) for t in range(n-1)])



"""
.. module:: mcmc
   :synopsis: MCMC sampler
.. moduleauthor:: Adam Moss <adam.moss@nottingham.ac.uk>
Performs MCMC to obtain posterior samples
"""

from __future__ import print_function
from __future__ import division

import os
import glob

import numpy as np
from getdist.mcsamples import MCSamples
from getdist.chains import chainFiles

from nnest.sampler import Sampler


class MCMCSampler(Sampler):

    def __init__(self,
                 x_dim,
                 loglike,
                 transform=None,
                 append_run_num=True,
                 run_num=None,
                 hidden_dim=128,
                 num_slow=0,
                 num_derived=0,
                 batch_size=100,
                 flow='nvp',
                 num_blocks=5,
                 num_layers=2,
                 log_dir='logs/test'):

        self.sampler = 'mcmc'

        super(MCMCSampler, self).__init__(x_dim, loglike, transform=transform, append_run_num=append_run_num,
                                          run_num=run_num, hidden_dim=hidden_dim, num_slow=num_slow, 
                                          num_derived=num_derived, batch_size=batch_size, flow=flow,
                                          num_blocks=num_blocks, num_layers=num_layers, log_dir=log_dir)

    def _init_samples(self, mcmc_steps=5000, mcmc_batch_size=5, ignore_rows=0.3):
        u = 2 * (np.random.uniform(size=(mcmc_batch_size, self.x_dim)) - 0.5)
        v = self.transform(u)
        logl = self.loglike(v)
        samples = []
        likes = []
        for i in range(mcmc_steps):
            du = np.random.standard_normal(u.shape) * 0.1
            u_prime = u + du
            v_prime = self.transform(u_prime)
            log_ratio_1 = np.zeros(mcmc_batch_size)
            prior = np.logical_or(np.abs(u) > 1, np.abs(u_prime) > 1)
            idx = np.where([np.any(p) for p in prior])
            log_ratio_1[idx] = -np.inf
            rnd_u = np.random.rand(mcmc_batch_size)
            ratio = np.clip(np.exp(log_ratio_1), 0, 1)
            mask = (rnd_u < ratio).astype(int)
            logl_prime = np.full(mcmc_batch_size, logl)
            for idx, im in enumerate(mask):
                if im:
                    lp = self.loglike(v_prime[idx])
                    if lp >= logl[idx]:
                        logl_prime[idx] = lp
                    elif rnd_u[idx] < np.clip(np.exp(lp - logl[idx]), 0, 1):
                        logl_prime[idx] = lp
                    else:
                        mask[idx] = 0
            m = mask[:, None]
            u = u_prime * m + u * (1 - m)
            v = v_prime * m + v * (1 - m)
            logl = logl_prime * mask + logl * (1 - mask)
            samples.append(v)
            likes.append(logl)
        samples = np.transpose(np.array(samples), axes=[1, 0, 2])
        loglikes = -np.transpose(np.array(likes), axes=[1, 0])
        weights = np.ones(loglikes.shape)
        self._chain_stats(samples)
        self._save_samples(samples, weights, loglikes)
        names = ['p%i' % i for i in range(int(self.x_dim))]
        labels = [r'x_%i' % i for i in range(int(self.x_dim))]
        files = chainFiles(os.path.join(self.logs['chains'], 'chain'), first_chain=1, last_chain=mcmc_batch_size)
        mc = MCSamples(self.logs['chains'], names=names, labels=labels, ignore_rows=ignore_rows)
        mc.readChains(files)
        return mc

    def _read_samples(self, fileroot, match='', ignore_rows=0.3):
        names = ['p%i' % i for i in range(int(self.num_params))]
        labels = [r'x_%i' % i for i in range(int(self.num_params))]
        if match:
            files = glob.glob(os.path.join(fileroot, match))
        else:
            files = chainFiles(fileroot)
        mc = MCSamples(fileroot, names=names, labels=labels, ignore_rows=ignore_rows)
        mc.readChains(files)
        return mc

    def run(
            self,
            train_iters=200,
            mcmc_steps=5000,
            bootstrap_iters=1,
            bootstrap_mcmc_steps=5000,
            bootstrap_fileroot='',
            bootstrap_match='',
            bootstrap_batch_size=5,
            alpha=0,
            single_thin=1,
            ignore_rows=0.3):

        if alpha == 0.0:
            alpha = 1 / self.x_dim ** 0.5

        if self.log:
            self.logger.info('Alpha [%5.4f]' % (alpha))

        for t in range(bootstrap_iters):

            if t == 0:
                if bootstrap_fileroot:
                    mc = self._read_samples(bootstrap_fileroot, match=bootstrap_match, ignore_rows=ignore_rows)
                else:
                    mc = self._init_samples(mcmc_steps=bootstrap_mcmc_steps, mcmc_batch_size=bootstrap_batch_size,
                                            ignore_rows=ignore_rows)
            else:
                samples, likes, latent, scale, nc = self.trainer.sample(
                    loglike=self.loglike, transform=transform,
                    mcmc_steps=bootstrap_mcmc_steps, alpha=alpha, dynamic=False, show_progress=True)
                samples = transform(samples)
                self._chain_stats(samples)
                loglikes = -np.array(likes)
                weights = np.ones(loglikes.shape)
                mc = MCSamples(samples=[samples[0]], weights=[weights[0]], loglikes=[loglikes[0]],
                               ignore_rows=ignore_rows)

            samples = mc.makeSingleSamples(single_thin=single_thin)
            samples = samples[:, :self.x_dim]
            mean = np.mean(samples, axis=0)
            std = np.std(samples, axis=0)
            samples = (samples - mean) / std
            self.trainer.train(samples, max_iters=train_iters, noise=-1)

            def transform(x):
                return x * std + mean

        samples, likes, latent, scale, nc = self.trainer.sample(
            loglike=self.loglike, transform=transform,
            mcmc_steps=mcmc_steps, alpha=alpha, dynamic=False, show_progress=True,
            out_chain=os.path.join(self.logs['chains'], 'chain'))
        samples = transform(samples)
        self._chain_stats(samples)



"""
.. module:: mcmc
   :synopsis: Sampler base class
.. moduleauthor:: Adam Moss <adam.moss@nottingham.ac.uk>
"""

from __future__ import print_function
from __future__ import division

import os
import json

import numpy as np

from nnest.trainer import Trainer
from nnest.utils.logger import create_logger, make_run_dir
from nnest.utils.evaluation import acceptance_rate, effective_sample_size, mean_jump_distance


class Sampler(object):

    def __init__(self,
                 x_dim,
                 loglike,
                 transform=None,
                 append_run_num=True,
                 run_num=None,
                 hidden_dim=128,
                 num_slow=0,
                 num_derived=0,
                 batch_size=100,
                 flow='nvp',
                 num_blocks=5,
                 num_layers=2,
                 log_dir='logs/test'
                 ):

        self.x_dim = x_dim
        self.num_params = x_dim + num_derived

        def safe_loglike(x):
            if len(x.shape) == 1:
                assert x.shape[0] == self.x_dim
                x = np.expand_dims(x, 0)
            logl = loglike(x)
            if len(logl.shape) == 0:
                logl = np.expand_dims(logl, 0)
            logl[np.logical_not(np.isfinite(logl))] = -1e100
            return logl

        self.loglike = safe_loglike

        if transform is None:
            self.transform = lambda x: x
        else:
            def safe_transform(x):
                if len(x.shape) == 1:
                    assert x.shape[0] == self.x_dim
                    x = np.expand_dims(x, 0)
                return transform(x)
            self.transform = safe_transform

        self.use_mpi = False
        try:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.mpi_size = self.comm.Get_size()
            self.mpi_rank = self.comm.Get_rank()
            if self.mpi_size > 1:
                self.use_mpi = True
        except:
            self.mpi_size = 1
            self.mpi_rank = 0

        self.log = not self.use_mpi or (self.use_mpi and self.mpi_rank == 0)

        args = locals()
        args.update(vars(self))

        if self.log:
            self.logs = make_run_dir(log_dir, run_num, append_run_num= append_run_num)
            log_dir = self.logs['run_dir']
            self._save_params(args)
        else:
            log_dir = None
                
        self.logger = create_logger(__name__)

        self.trainer = Trainer(
                x_dim,
                hidden_dim,
                nslow=num_slow,
                batch_size=batch_size,
                flow=flow,
                num_blocks=num_blocks,
                num_layers=num_layers,
                log_dir=log_dir,
                log=self.log)

    def _save_params(self, my_dict):
        my_dict = {k: str(v) for k, v in my_dict.items()}
        with open(os.path.join(self.logs['info'], 'params.txt'), 'w') as f:
            json.dump(my_dict, f, indent=4)

    def _chain_stats(self, samples, mean=None, std=None):
        acceptance = acceptance_rate(samples)
        if mean is None:
            mean = np.mean(np.reshape(samples, (-1, samples.shape[2])), axis=0)
        if std is None:
            std = np.std(np.reshape(samples, (-1, samples.shape[2])), axis=0)
        ess = effective_sample_size(samples, mean, std)
        jump_distance = mean_jump_distance(samples)
        self.logger.info(
            'Acceptance [%5.4f] min ESS [%5.4f] max ESS [%5.4f] jump distance [%5.4f]' %
            (acceptance, np.min(ess), np.max(ess), jump_distance))
        return acceptance, ess, jump_distance

    def _save_samples(self, samples, weights, logl, min_weight=1e-30, outfile='chain'):
        if len(samples.shape) == 2:
            with open(os.path.join(self.logs['chains'], outfile + '.txt'), 'w') as f:
                for i in range(samples.shape[0]):
                    f.write("%.5E " % max(weights[i], min_weight))
                    f.write("%.5E " % -logl[i])
                    f.write(" ".join(["%.5E" % v for v in samples[i, :]]))
                    f.write("\n")
        elif len(samples.shape) == 3:
            for ib in range(samples.shape[0]):
                with open(os.path.join(self.logs['chains'], outfile + '_%s.txt' % (ib+1)), 'w') as f:
                    for i in range(samples.shape[1]):
                        f.write("%.5E " % max(weights[ib, i], min_weight))
                        f.write("%.5E " % -logl[ib, i])
                        f.write(" ".join(["%.5E" % v for v in samples[ib, i, :]]))
                        f.write("\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

setup(
    name = "nnest",
    version = "0.1.1",
    description = "Neural network nested sampling",
    author = "Adam Moss",
    author_email = "adam.moss@nottingham.ac.uk",
    maintainer = "Adam Moss",
    maintainer_email = "adam.moss@nottingham.ac.uk",
    url = "https://github.com/adammoss/nnest/",
    license = "MIT",
    packages = find_packages(),
    provides = ["nnest"],
    requires = ["torch", "tensorflow", "tensorboardX", "numpy", "scipy", "matplotlib", "pandas",
                "scikitlearn", "tqdm", "pillow"],
)
